"""
Matchers for collections.
"""
from __future__ import absolute_import

try:
    import collections.abc
    abc = collections.abc
except ImportError:
    import collections
    abc = collections

import inspect

from calleee.base import BaseMatcher
from calleee.general import Any
from calleee.types import InstanceOf


__all__ = [
    'Iterable', 'Generator',
    'Sequence', 'List', 'Set',
    'Mapping', 'Dict', 'OrderedDict',
]


class CollectionMatcher(BaseMatcher):
    """Base class for collections' matchers.
    This class shouldn't be used directly.
    """
    #: Collection class to match.
    #: Must be overridden in subclasses.
    CLASS = None

    def __init__(self, of=None):
        """
        :param of: Optional matcher for the elements,
                   or the expected type of the elements.
        """
        assert self.CLASS, "must specify collection type to match"
        self.of = self._validate_argument(of)

    def _validate_argument(self, arg):
        """Validate a type or matcher argument to the constructor."""
        if arg is None:
            return arg

        if isinstance(arg, type):
            return InstanceOf(arg)
        if not isinstance(arg, BaseMatcher):
            raise TypeError(
                "argument of %s can be a type or a matcher (got %r)" % (
                    self.__class__.__name__, type(arg)))

        return arg

    def match(self, value):
        if not isinstance(value, self.CLASS):
            return False
        if self.of is not None:
            return all(self.of == item for item in value)
        return True

    def __repr__(self):
        """Return a readable representation of the matcher.
        Used mostly for AssertionError messages in failed tests.

        Example::

            <List[<Integer>]>
        """
        of = "" if self.of is None else "[%r]" % (self.of,)
        return "<%s%s>" % (self.__class__.__name__, of)


class Iterable(CollectionMatcher):
    """Matches any iterable."""

    CLASS = abc.Iterable

    def __init__(self):
        # Unfortunately, we can't allow an ``of`` argument to this matcher.
        #
        # An otherwise unspecified iterable can't be iterated upon
        # more than once safely, because it could be a one-off iterable
        # (e.g. generator comprehension) that's exhausted after a single pass.
        #
        # Thus the sole act of checking the element types would alter
        # the object we're trying to match, and potentially cause all sorts
        # of unexpected behaviors (e.g. tests passing/failing depending on
        # the order of assertions).
        #
        super(Iterable, self).__init__(of=None)


class Generator(BaseMatcher):
    r"""Matches an iterable that's a generator.

    A generator can be a generator expression ("comprehension")
    or an invocation of a generator function (one that ``yield``\ s objects).

    .. note::

        To match a *generator function* itself, you should use the
        :class:`~calleee.functions.GeneratorFunction` matcher instead.
    """
    def match(self, value):
        return inspect.isgenerator(value)

    def __repr__(self):
        return "<Generator>"


# Ordinary collections

class Sequence(CollectionMatcher):
    """Matches a sequence of given items.

    A sequence is an iterable that has a length and can be indexed.
    """
    CLASS = abc.Sequence


class List(CollectionMatcher):
    """Matches a :class:`list` of given items."""

    CLASS = list


class Set(CollectionMatcher):
    """Matches a :class:`set` of given items."""

    CLASS = abc.Set


# TODO: Tuple matcher, with of= that accepts a tuple of matchers
# so that tuple elements can be also matched on


# Mappings

class MappingMatcher(CollectionMatcher):
    """Base class for mapping matchers.
    This class shouldn't be used directly.
    """
    #: Mapping class to match.
    #: Must be overridden in subclasses.
    CLASS = None

    def __init__(self, *args, **kwargs):
        """Constructor can be invoked either with parameters described below
        (given as keyword arguments), or with two positional arguments:
        matchers/types for dictionary keys & values::

            Dict(String(), int)  # dict mapping strings to ints

        :param keys: Matcher for dictionary keys.
        :param values: Matcher for dictionary values.
        :param of: Matcher for dictionary items, or a tuple of matchers
                   for keys & values, e.g. ``(String(), Integer())``.
                   Cannot be provided if either ``keys`` or ``values``
                   is also passed.

        """
        assert self.CLASS, "must specify mapping type to match"
        self._initialize(*args, **kwargs)

    def _initialize(self, *args, **kwargs):
        """Initiaize the mapping matcher with constructor arguments."""
        self.items = None
        self.keys = None
        self.values = None

        if args:
            if len(args) != 2:
                raise TypeError("expected exactly two positional arguments, "
                                "got %s" % len(args))
            if kwargs:
                raise TypeError(
                    "expected positional or keyword arguments, not both")

            # got positional arguments only
            self.keys, self.values = map(self._validate_argument, args)
        elif kwargs:
            has_kv = 'keys' in kwargs and 'values' in kwargs
            has_of = 'of' in kwargs
            if not (has_kv or has_of):
                raise TypeError("expected keys/values or items matchers, "
                                "but got: %s" % list(kwargs.keys()))
            if has_kv and has_of:
                raise TypeError(
                    "expected keys & values, or items matchers, not both")

            if has_kv:
                # got keys= and values= matchers
                self.keys = self._validate_argument(kwargs['keys'])
                self.values = self._validate_argument(kwargs['values'])
            else:
                # got of= matcher, which can be a tuple of matchers,
                # or a single matcher for dictionary items
                of = kwargs['of']
                if isinstance(of, tuple):
                    try:
                        # got of= as tuple of matchers
                        self.keys, self.values = \
                            map(self._validate_argument, of)
                    except ValueError:
                        raise TypeError(
                            "of= tuple has to be a pair of matchers/types" % (
                                self.__class__.__name__,))
                else:
                    # got of= as a single matcher
                    self.items = self._validate_argument(of)

    def match(self, value):
        if not isinstance(value, self.CLASS):
            return False

        if self.items is not None:
            return all(self.items == i for i in value.items())
        if self.keys is not None and self.values is not None:
            return all(self.keys == k and self.values == v
                       for k, v in value.items())

        return True

    def __repr__(self):
        """Return a readable representation of the matcher
        Used mostly for AssertionError messages in failed tests.

        Example::

            <Dict[<String> => <Any>]>
        """
        of = ""

        if self.items is not None:
            of = "[%r]" % self.items

        if self.keys is not None or self.values is not None:
            keys = repr(Any() if self.keys is None else self.keys)
            values = repr(Any() if self.values is None else self.values)
            of = "[%s => %s]" % (keys, values)

        return "<%s%s>" % (self.__class__.__name__, of)


class Mapping(MappingMatcher):
    """Matches a mapping of given items."""

    CLASS = abc.Mapping


class Dict(MappingMatcher):
    """Matches a dictionary (:class:`dict`) of given items."""

    CLASS = dict


class OrderedDict(MappingMatcher):
    """Matches an ordered dictionary (:class:`collections.OrderedDict`)
    of given items.
    """
    CLASS = collections.OrderedDict



"""
Base classes for argument matchers.
"""
import inspect
from operator import itemgetter

from calleee._compat import IS_PY3, metaclass


__all__ = [
    'Matcher',
    'Eq', 'Is', 'IsNot',
    'Not', 'And', 'Or', 'Either', 'OneOf', 'Xor',
]


class BaseMatcherMetaclass(type):
    """Metaclass for :class:`BaseMatcher`."""

    #: What __magic__ methods of :class:`BaseMatcher`
    #: can be overriden by user-defined subclasses.
    #:
    #: Any method not on this list can only be overridden by classes defined
    #: within this module. This prevents users from accidentally interfering
    #: with fundamental matcher functionality while writing their own matchers.
    #
    #: The names are given without the leading or trailing underscores.
    #:
    USER_OVERRIDABLE_MAGIC_METHODS = ('init', 'repr')

    def __new__(meta, classname, bases, dict_):
        """Create a new matcher class."""
        meta._validate_class_definition(classname, bases, dict_)
        return super(BaseMatcherMetaclass, meta) \
            .__new__(meta, classname, bases, dict_)

    @classmethod
    def _validate_class_definition(meta, classname, bases, dict_):
        """Ensure the matcher class definition is acceptable.
        :raise RuntimeError: If there is a problem
        """
        # let the BaseMatcher class be created without hassle
        if meta._is_base_matcher_class_definition(classname, dict_):
            return

        # ensure that no important magic methods are being overridden
        for name, member in dict_.items():
            if not (name.startswith('__') and name.endswith('__')):
                continue

            # check if it's not a whitelisted magic method name
            name = name[2:-2]
            if not name:
                continue  # unlikely case of a ``____`` function
            if name not in meta._list_magic_methods(BaseMatcher):
                continue
            if name in meta.USER_OVERRIDABLE_MAGIC_METHODS:
                continue

            # non-function attributes, like __slots__, are harmless
            if not inspect.isfunction(member):
                continue

            # classes in this very module are exempt, since they define
            # the very behavior of matchers we want to protect
            if member.__module__ == __name__:
                continue

            raise RuntimeError(
                "matcher class %s cannot override the __%s__ method" % (
                    classname, name))

    @classmethod
    def _is_base_matcher_class_definition(meta, classname, dict_):
        """Checks whether given class name and dictionary
        define the :class:`BaseMatcher`.
        """
        if classname != 'BaseMatcher':
            return False
        methods = list(filter(inspect.isfunction, dict_.values()))
        return methods and all(m.__module__ == __name__ for m in methods)

    @classmethod
    def _list_magic_methods(meta, class_):
        """Return names of magic methods defined by a class.
        :return: Iterable of magic methods, each w/o the ``__`` prefix/suffix
        """
        return [
            name[2:-2] for name, member in class_.__dict__.items()
            if len(name) > 4 and name.startswith('__') and name.endswith('__')
            and inspect.isfunction(member)
        ]

    # TODO: consider making matcher classes interchangeable with matcher
    # objects created w/o ctor args, i.e. making Integer and Integer()
    # equivalent; it'd require this metaclass to implement the magic methods
    # from BaseMatcher and something better than `isinstance(x, BaseMatcher)`


@metaclass(BaseMatcherMetaclass)
class BaseMatcher(object):
    """Base class for all argument matchers.

    This class shouldn't be used directly by the clients.
    To create custom matchers, inherit from :class:`Matcher` instead.
    """
    __slots__ = ()

    def match(self, value):
        raise NotImplementedError("matching not implemented")

    def __repr__(self):
        return "<unspecified matcher>"

    def __eq__(self, other):
        if isinstance(other, BaseMatcher):
            raise TypeError(
                "incorrect use of matcher object as a value to match on")
        return self.match(other)

    # TODO: make matcher objects callable

    def __invert__(self):
        return Not(self)

    def __and__(self, other):
        matchers = other._matchers if isinstance(other, And) else [other]
        return And(self, *matchers)

    def __or__(self, other):
        matchers = other._matchers if isinstance(other, Or) else [other]
        return Or(self, *matchers)

    def __xor__(self, other):
        matchers = other._matchers if isinstance(other, Either) else [other]
        return Either(self, *matchers)


class Matcher(BaseMatcher):
    """Base class for custom (user-defined) argument matchers.

    To create a custom matcher, simply inherit from this class
    and implement the :meth:`match` method.

    If the matcher is more complicated (e.g. parametrized),
    you may also  want to provide a :meth:`__repr__` method implementation
    for better error messages.
    """
    def __repr__(self):
        r"""Provides a default ``repr``\ esentation for custom matchers.

        This representation will include matcher class name
        and the values of its public attributes.
        If that's insufficient, consider overriding this method.
        """
        args = ""

        # check if the matcher class has a parametrized constructor
        has_argful_ctor = False
        if '__init__' in self.__class__.__dict__:
            argnames, vargargs, kwargs, _ = inspect.getargspec(
                self.__class__.__init__)
            has_argful_ctor = bool(argnames[1:] or vargargs or kwargs)

        # if so, then it probably means it has some interesting state
        # in its attributes which we can include in the default representation
        if has_argful_ctor:
            # TODO: __getstate__ instead of __dict__?
            fields = [(name, value) for name, value in self.__dict__.items()
                      if not name.startswith('_')]
            if fields:
                def repr_value(value):
                    """Safely represent a value as an ASCII string."""
                    if isinstance(value, bytes):
                        value = value.decode('ascii', 'ignore')
                    if not IS_PY3 and isinstance(value, unicode):
                        value = value.encode('ascii', 'replace')
                        value = str(value)
                    return repr(value)

                fields.sort(key=itemgetter(0))
                args = "(%s)" % ", ".join(
                    "%s=%s" % (name, repr_value(value)[:32])
                    for name, value in fields)
            else:
                args = "(...)"

        return "<%s%s>" % (self.__class__.__name__, args)


# Special cases around equality/identity

class Eq(BaseMatcher):
    """Matches a value exactly using the equality (``==``) operator.

    This is already the default mode of operation for ``assert_called_with``
    methods on mocks, making this matcher redundant in most situations::

        mock_foo.assert_called_with(bar)
        mock_foo.assert_called_with(Eq(bar))  # equivalent

    In very rare and specialized cases, however, if the **tested code** treats
    `calleee` matcher objects in some special way, using :class:`Eq` may be
    necessary.

    Those situations shouldn't generally arise outside of writing tests
    for code that is itself a test library or helper.
    """
    def __init__(self, value):
        """:param value: Value to match against"""
        self.value = value

    def match(self, value):
        return self.value == value

    def __eq__(self, other):
        return self.match(other)

    def __repr__(self):
        # This representation matches the format of comparison operators
        # (such as :class:`Less`) defined in the ``.operators`` module.
        return "<... == %r>" % (self.value,)


class Is(BaseMatcher):
    """Matches a value using the identity (``is``) operator."""

    def __init__(self, value):
        self.value = value

    def match(self, value):
        return value is self.value

    def __eq__(self, other):
        return self.match(other)

    def __repr__(self):
        # This representation matches the format of comparison operators
        # (such as :class:`Less`) defined in the ``.operators`` module.
        return "<... is %r>" % (self.value,)


class IsNot(BaseMatcher):
    """Matches a value using the negated identity (``is not``) operator."""

    def __init__(self, value):
        self.value = value

    def match(self, value):
        return value is not self.value

    def __eq__(self, other):
        return self.match(other)

    def __repr__(self):
        # This representation matches the format of comparison operators
        # (such as :class:`Less`) defined in the ``.operators`` module.
        return "<... is not %r>" % (self.value,)


# Logical combinators for matchers

class Not(BaseMatcher):
    """Negates given matcher.

    :param matcher: Matcher object to negate the semantics of
    """
    def __init__(self, matcher):
        assert isinstance(matcher, BaseMatcher), "Not() expects a matcher"
        self._matcher = matcher

    def match(self, value):
        return not self._matcher.match(value)

    def __repr__(self):
        return "not %r" % (self._matcher,)

    def __invert__(self):
        return self._matcher

    def __and__(self, other):
        # convert (~a) & (~b) into ~(a | b) which is one operation less
        # but still equivalent as per de Morgan laws
        if isinstance(other, Not):
            return Not(self.matcher | other.matcher)
        return super(Not, self).__and__(other)

    def __or__(self, other):
        # convert (~a) | (~b) into ~(a & b) which is one operation less
        # but still equivalent as per de Morgan laws
        if isinstance(other, Not):
            return Not(self.matcher & other.matcher)
        return super(Not, self).__or__(other)


class And(BaseMatcher):
    """Matches the argument only if all given matchers do."""

    def __init__(self, *matchers):
        assert matchers, "And() expects at least one matcher"
        assert all(isinstance(m, BaseMatcher)
                   for m in matchers), "And() expects matchers"
        self._matchers = list(matchers)

    # TODO: coalesce a & b & c into single And(a, b, c)

    def match(self, value):
        return all(matcher.match(value) for matcher in self._matchers)

    def __repr__(self):
        return "<%s>" % " and ".join(map(repr, self._matchers))


class Or(BaseMatcher):
    """Matches the argument only if at least one given matcher does."""

    def __init__(self, *matchers):
        assert matchers, "Or() expects at least one matcher"
        assert all(isinstance(m, BaseMatcher)
                   for m in matchers), "Or() expects matchers"
        self._matchers = list(matchers)

    # TODO: coalesce a | b | c into single Or(a, b, c)

    def match(self, value):
        return any(matcher.match(value) for matcher in self._matchers)

    def __repr__(self):
        return "<%s>" % " or ".join(map(repr, self._matchers))


class Either(BaseMatcher):
    """Matches the argument only if some (but not all) of given matchers do.

    .. versionadded:: 0.3
    """
    def __init__(self, *matchers):
        assert len(matchers) >= 2, "Either() expects at least two matchers"
        assert all(isinstance(m, BaseMatcher)
                   for m in matchers), "Either() expects matchers"
        self._matchers = list(matchers)

    def match(self, value):
        any_matches = bool(self._matchers[0].match(value))
        for matcher in self._matchers[1:]:
            is_match = bool(matcher.match(value))
            if is_match != any_matches:
                return True
            any_matches |= is_match
        return False

    def __repr__(self):
        return "<%s>" % " xor ".join(map(repr, self._matchers))

#: Alias for :class:`Either`.
OneOf = Either
#: Alias for :class:`Either`.
Xor = Either



"""
Compatibility shims for different Python versions.
"""
from __future__ import absolute_import

try:
    import asyncio
except ImportError:
    asyncio = None
import inspect
import sys


__all__ = [
    'asyncio',
    'IS_PY3',
    'STRING_TYPES', 'casefold',
    'metaclass',
    'getargspec',
]


IS_PY3 = sys.version_info[0] == 3

STRING_TYPES = (str,) if IS_PY3 else (basestring,)
casefold = getattr(str, 'casefold', None) or (lambda s: s.lower())


class MetaclassDecorator(object):
    """Decorator for creating a class through a metaclass.

    Unlike ``__metaclass__`` attribute from Python 2, or ``metaclass=`` keyword
    argument from Python 3, the ``@metaclass`` decorator works with both
    versions of the language.

    Example::

        @metaclass(MyMetaclass)
        class MyClass(object):
            pass
    """
    def __init__(self, meta):
        if not issubclass(meta, type):
            raise TypeError(
                "expected a metaclass, got %s instead" % type(meta).__name__)
        self.metaclass = meta

    def __call__(self, cls):
        """Apply the decorator to given class.

        This recreates the class using the previously supplied metaclass.
        """
        # Copyright (c) Django Software Foundation and individual contributors.
        # All rights reserved.

        original_dict = cls.__dict__.copy()
        original_dict.pop('__dict__', None)
        original_dict.pop('__weakref__', None)

        slots = original_dict.get('__slots__')
        if slots is not None:
            if isinstance(slots, str):
                slots = [slots]
            for slot in slots:
                original_dict.pop(slot)

        return self.metaclass(cls.__name__, cls.__bases__, original_dict)

metaclass = MetaclassDecorator
del MetaclassDecorator


def getargspec(obj):
    """Portable version of inspect.getargspec().

    Necessary because the original is no longer available
    starting from Python 3.6.

    :return: 4-tuple of (argnames, varargname, kwargname, defaults)

    Note that distinction between positional-or-keyword and keyword-only
    parameters will be lost, as the original getargspec() doesn't honor it.
    """
    try:
        return inspect.getargspec(obj)
    except AttributeError:
        pass  # we let a TypeError through

    # translate the signature object back into the 4-tuple
    argnames = []
    varargname, kwargname = None, None
    defaults = []
    for name, param in inspect.signature(obj):
        if param.kind == inspect.Parameter.VAR_POSITIONAL:
            varargname = name
        elif param.kind == inspect.Parameter.VAR_KEYWORD:
            kwargname = name
        else:
            argnames.append(name)
            if param.default is not inspect.Parameter.empty:
                defaults.append(param.default)
    defaults = defaults or None

    return argnames, varargname, kwargname, defaults



"""
General matchers.

These don't belong to any broader category, and include matchers for common
Python objects, like functions or classes.
"""
import inspect

from calleee._compat import IS_PY3, STRING_TYPES
from calleee.base import BaseMatcher


__all__ = [
    'Any', 'Matching', 'ArgThat', 'Captor',
]


# TODO: introduce custom exception types rather than using built-ins

# TODO: matchers for positional & keyword arguments,
# e.g. *Args(Integer(), min=1, max=10), **Kwargs(foo=String(), bar=Float())


class Any(BaseMatcher):
    """Matches any object."""

    def match(self, value):
        return True

    def __repr__(self):
        return "<Any>"


class Matching(BaseMatcher):
    """Matches an object that satisfies given predicate."""

    MAX_DESC_LENGTH = 32

    def __init__(self, predicate, desc=None):
        """
        :param predicate: Callable taking a single argument
                          and returning True or False
        :param desc: Optional description of the predicate.
                     This will be displayed as a part of the error message
                     on failed assertion.
        """
        if not callable(predicate):
            raise TypeError(
                "Matching requires a predicate, got %r" % (predicate,))

        self.predicate = predicate
        self.desc = self._validate_desc(desc)

    def _validate_desc(self, desc):
        """Validate the predicate description."""
        if desc is None:
            return desc

        if not isinstance(desc, STRING_TYPES):
            raise TypeError(
                "predicate description for Matching must be a string, "
                "got %r" % (type(desc),))

        # Python 2 mandates __repr__ to be an ASCII string,
        # so if Unicode is passed (usually due to unicode_literals),
        # it should be ASCII-encodable.
        if not IS_PY3 and isinstance(desc, unicode):
            try:
                desc = desc.encode('ascii', errors='strict')
            except UnicodeEncodeError:
                raise TypeError("predicate description must be "
                                "an ASCII string in Python 2")

        return desc

    def match(self, value):
        # Note that any possible exceptions from ``predicate``
        # are intentionally let through, to make it easier to diagnose errors
        # than a plain "no match" response would.
        return bool(self.predicate(value))
        # TODO: translate exceptions from the predicate into our own
        # exception type to not clutter user-visible stracktraces with our code

    def __repr__(self):
        """Return a representation of the matcher."""
        name = getattr(self.predicate, '__name__', None)
        desc = self.desc

        # When no user-provided description is available,
        # use function's own name or even its repr().
        if desc is None:
            # If not a lambda function, we can probably make the representation
            # more readable by showing just the function's own name.
            if name and name != '<lambda>':
                # Where possible, make it a fully qualified name, including
                # the module path. This is either on Python 3.3+
                # (via __qualname__), or when the predicate is
                # a standalone function (not a method).
                qualname = getattr(self.predicate, '__qualname__', name)
                is_method = inspect.ismethod(self.predicate) or \
                    isinstance(self.predicate, staticmethod)
                if qualname != name or not is_method:
                    # Note that this shows inner functions (those defined
                    # locally inside other functions) as if they were global
                    # to the module.
                    # This is why we use colon (:) as separator here, as to not
                    # suggest this is an evaluatable identifier.
                    name = '%s:%s' % (self.predicate.__module__, qualname)
            else:
                # For lambdas and other callable objects,
                # we'll just default to the Python repr().
                name = None
        else:
            # Quote and possibly ellipsize the provided description.
            if len(desc) > self.MAX_DESC_LENGTH:
                ellipsis = '...'
                desc = desc[:self.MAX_DESC_LENGTH - len(ellipsis)] + ellipsis
            desc = '"%s"' % desc

        return "<Matching %s>" % (desc or name or repr(self.predicate))

ArgThat = Matching


class Captor(BaseMatcher):
    """Argument captor.

    You can use :class:`Captor` to "capture" the original argument
    that the mock was called with, and perform custom assertions on it.

    Example::

        captor = Captor()
        mock_foo.assert_called_with(captor)

        # captured value is available as the `arg` attribute
        self.assertEquals(captor.arg.some_method(), 42)
        self.assertEquals(captor.arg.some_other_method(), "foo")

    .. versionadded:: 0.2
    """
    __slots__ = ('matcher', 'value')

    def __init__(self, matcher=None):
        """
        :param matcher: Optional matcher to validate the argument against
                        before it's captured
        """
        if matcher is None:
            matcher = Any()

        if not isinstance(matcher, BaseMatcher):
            raise TypeError("expected a matcher, got %r" % (type(matcher),))
        if isinstance(matcher, Captor):
            raise TypeError("cannot pass a captor to another captor")

        self.matcher = matcher

    def has_value(self):
        """Returns whether the :class:`Captor` has captured a value."""
        return hasattr(self, 'value')

    @property
    def arg(self):
        """The captured argument value."""
        if not self.has_value():
            raise ValueError("no value captured")
        return self.value

    def match(self, value):
        if self.has_value():
            raise ValueError("a value has already been captured")

        if not self.matcher.match(value):
            return False
        self.value = value
        return True

    def __repr__(self):
        """Return a representation of the captor."""
        return "<Captor %r%s>" % (self.matcher,
                                  " (*)" if self.has_value() else "")



"""
calleee
"""
__version__ = "0.4.0"
__description__ = "Argument matchers for unittest.mock"
__author__ = "Karol Kuczmarski"
__license__ = "BSD"


from calleee.attributes import Attrs, Attr, HasAttrs, HasAttr
from calleee.base import \
    And, Either, Eq, Is, IsNot, OneOf, Or, Matcher, Not, Xor
from calleee.collections import \
    Dict, Generator, Iterable, List, Mapping, OrderedDict, Sequence, Set
from calleee.general import Any, ArgThat, Captor, Matching
from calleee.functions import \
    Callable, CoroutineFunction, Function, GeneratorFunction
from calleee.numbers import (Complex, Float, Fraction, Int, Integer, Integral,
                            Long, Number, Rational, Real)
from calleee.objects import Bytes, Coroutine, FileLike
from calleee.operators import (
    Contains,
    Ge, Greater, GreaterOrEqual, GreaterOrEqualTo, GreaterThan, Gt,
    In,
    Le, Less, LessOrEqual, LessOrEqualTo, LessThan,
    Longer, LongerOrEqual, LongerOrEqualTo, LongerThan, Lt,
    Shorter, ShorterOrEqual, ShorterOrEqualTo, ShorterThan)
from calleee.strings import \
    EndsWith, Glob, Regex, StartsWith, String, Unicode
from calleee.types import InstanceOf, IsA, SubclassOf, Inherits, Type, Class


__all__ = [
    'Matcher',
    'Eq', 'Is', 'IsNot',
    'Not', 'And', 'Or', 'Either', 'OneOf', 'Xor',

    'Attrs', 'Attr', 'HasAttrs', 'HasAttr',

    'Iterable', 'Generator',
    'Sequence', 'List', 'Set',
    'Mapping', 'Dict', 'OrderedDict',

    'Any', 'Matching', 'ArgThat', 'Captor',

    'Callable', 'Function', 'GeneratorFunction',
    'CoroutineFunction',

    'Number',
    'Complex', 'Real', 'Float', 'Rational', 'Fraction',
    'Integral', 'Integer', 'Int', 'Long',

    'Bytes',
    'Coroutine',
    'FileLike',

    'Less', 'LessThan', 'Lt',
    'LessOrEqual', 'LessOrEqualTo', 'Le',
    'Greater', 'GreaterThan', 'Gt',
    'GreaterOrEqual', 'GreaterOrEqualTo', 'Ge',
    'Shorter', 'ShorterThan', 'ShorterOrEqual', 'ShorterOrEqualTo',
    'Longer', 'LongerThan', 'LongerOrEqual', 'LongerOrEqualTo',
    'Contains', 'In',

    'String', 'Unicode',
    'StartsWith', 'EndsWith', 'Glob', 'Regex',

    'InstanceOf', 'IsA', 'SubclassOf', 'Inherits', 'Type', 'Class',
]



"""
Attribute-based matchers.
"""
from itertools import starmap
from operator import itemgetter

from calleee.base import BaseMatcher, Eq


__all__ = [
    'Attrs', 'Attr', 'HasAttrs', 'HasAttr',
]


class Attrs(BaseMatcher):
    """Matches objects based on their attributes.

    To match successfully, the object needs to:

        * have all the attributes whose names were passed
          as positional arguments (regardless of their values)
        * have the attribute names/values that correspond exactly
          to keyword arguments' names and values

    Examples::

        Attrs('foo')  # `foo` attribute with any value
        Attrs('foo', 'bar')  # `foo` and `bar` attributes with any values
        Attrs(foo=42)  # `foo` attribute with value of 42
        Attrs(bar=Integer())  # `bar` attribute whose value is an integer
        Attrs('foo', bar='x')  # `foo` with any value, `bar` with value of 'x'
    """
    def __init__(self, *args, **kwargs):
        if not (args or kwargs):
            raise TypeError("%s() requires at least one argument" % (
                self.__class__.__name__,))

        self.attr_names = list(args)
        self.attr_dict = dict((k, v if isinstance(v, BaseMatcher) else Eq(v))
                              for k, v in kwargs.items())

    def match(self, value):
        for name in self.attr_names:
            # Can't use hasattr() here because it swallows *all* exceptions
            # from attribute access in Python 2.x, not just AttributeError.
            # More details: https://hynek.me/articles/hasattr/
            try:
                getattr(value, name)
            except AttributeError:
                return False

        for name, matcher in self.attr_dict.items():
            # Separately handle retrieving of the attribute value,
            # so that any stray AttributeErrors from the matcher itself
            # are correctly propagated.
            try:
                attrvalue = getattr(value, name)
            except AttributeError:
                return False
            if not matcher.match(attrvalue):
                return False

        return True

    def __repr__(self):
        """Return a representation of the matcher."""
        # get both the names-only and valued attributes and sort them by name
        sentinel = object()
        attrs = [(name, sentinel)
                 for name in self.attr_names] + list(self.attr_dict.items())
        attrs.sort(key=itemgetter(0))

        def attr_repr(name, value):
            # include the value with attribute name whenever necessary
            if value is sentinel:
                return name
            value = value.value if isinstance(value, Eq) else value
            return "%s=%r" % (name, value)

        return "<%s %s>" % (self.__class__.__name__,
                            " ".join(starmap(attr_repr, attrs)))


class Attr(Attrs):
    """Matches objects that have an attribute with given name and value,
    as given by a keyword argument.
    """
    def __init__(self, **kwargs):
        if not len(kwargs) == 1:
            raise TypeError("Attr() requires exactly one keyword argument")
        super(Attr, self).__init__(**kwargs)


class HasAttrs(Attrs):
    """Matches objects that have all of the specified attribute names,
    regardless of their values.
    """
    def __init__(self, *args):
        super(HasAttrs, self).__init__(*args)


class HasAttr(HasAttrs):
    """Matches object that have an attribute with given name,
    regardless of its value.
    """
    def __init__(self, name):
        super(HasAttr, self).__init__(name)



"""
Matchers based on Python operators.
"""
from __future__ import absolute_import

from numbers import Number
import operator

from calleee.base import BaseMatcher, Eq, Is, IsNot


__all__ = [
    # those are defined elsewhere but they fit in this module, too
    'Eq', 'Is', 'IsNot',

    'Less', 'LessThan', 'Lt',
    'LessOrEqual', 'LessOrEqualTo', 'Le',
    'Greater', 'GreaterThan', 'Gt',
    'GreaterOrEqual', 'GreaterOrEqualTo', 'Ge',

    'Shorter', 'ShorterThan', 'ShorterOrEqual', 'ShorterOrEqualTo',
    'Longer', 'LongerThan', 'LongerOrEqual', 'LongerOrEqualTo',

    'Contains', 'In',
]


class OperatorMatcher(BaseMatcher):
    """Matches values based on comparison operator and a reference object.
    This class shouldn't be used directly.
    """
    #: Operator function to use for comparing a value with a reference object.
    #: Must be overridden in subclasses.
    OP = None

    #: Transformation function to apply to given value before comparison.
    TRANSFORM = None

    def __init__(self, *args, **kwargs):
        """Accepts a single argument: the reference object to compare against.

        It can be passed either as a single positional parameter,
        or as a single keyword argument -- preferably with a readable name,
        for example::

            some_mock.assert_called_with(Number() & LessOrEqual(to=42))
        """
        assert self.OP, "must specify comparison operator to use"

        # check that we've received exactly one argument,
        # either positional or keyword
        argcount = len(args) + len(kwargs)
        if argcount != 1:
            raise TypeError("a single argument expected, got %s" % argcount)
        if len(args) > 1:
            raise TypeError(
                "at most one positional argument expected, got %s" % len(args))
        if len(kwargs) > 1:
            raise TypeError(
                "at most one keyword argument expected, got %s" % len(kwargs))

        # extract the reference object from arguments
        ref = None
        if args:
            ref = args[0]
        elif kwargs:
            _, ref = kwargs.popitem()

        #: Reference object to compare given values to.
        self.ref = ref

    def match(self, value):
        # Note that any possible exceptions from either ``TRANSFORM`` or ``OP``
        # are intentionally let through, to make it easier to diagnose errors
        # than a plain "no match" response would.
        if self.TRANSFORM is not None:
            value = self.TRANSFORM(value)
        return self.OP(value, self.ref)

    def __repr__(self):
        """Provide an universal representation of the matcher."""
        # Mapping from operator functions to their symbols in Python.
        #
        # There is no point in including ``operator.contains`` due to lack of
        # equivalent ``operator.in_``.
        # These are handled by membership matchers directly.
        operator_symbols = {
            operator.eq: '==',
            operator.ge: '>=',
            operator.gt: '>',
            operator.le: '<=',
            operator.lt: '<',
        }

        # try to get the symbol for the operator, falling back to Haskell-esque
        # "infix function" representation
        op = operator_symbols.get(self.OP)
        if op is None:
            # TODO: convert CamelCase to either snake_case or kebab-case
            op = '`%s`' % (self.__class__.__name__.lower(),)

        return "<%s %s %r>" % (self._get_placeholder_repr(), op, self.ref)

    def _get_placeholder_repr(self):
        """Return the placeholder part of matcher's ``__repr__``."""
        placeholder = '...'
        if self.TRANSFORM is not None:
            placeholder = '%s(%s)' % (self.TRANSFORM.__name__, placeholder)
        return placeholder


# Simple comparisons

# TODO: AlmostEq() for approximate comparisons with epsilon

class Less(OperatorMatcher):
    """Matches values that are smaller (as per ``<`` operator)
    than given object.
    """
    OP = operator.lt

LessThan = Less

Lt = Less


class LessOrEqual(OperatorMatcher):
    """Matches values that are smaller than,
    or equal to (as per ``<=`` operator), given object.
    """
    OP = operator.le

LessOrEqualTo = LessOrEqual

Le = LessOrEqual


class Greater(OperatorMatcher):
    """Matches values that are greater (as per ``>`` operator)
    than given object.
    """
    OP = operator.gt

GreaterThan = Greater

Gt = Greater


class GreaterOrEqual(OperatorMatcher):
    """Matches values that are greater than,
    or equal to (as per ``>=`` operator), given object.
    """
    OP = operator.ge

GreaterOrEqualTo = GreaterOrEqual

Ge = GreaterOrEqual


# Length comparisons

class LengthMatcher(OperatorMatcher):
    """Matches values based on their length, as compared to a reference.
    This class shouldn't be used directly.
    """
    TRANSFORM = len

    def __init__(self, *args, **kwargs):
        super(LengthMatcher, self).__init__(*args, **kwargs)

        # allow the reference to be either a numeric length or another sequence
        # TODO: remember at least the sequence type to make it impossible
        # e.g. to accidentally compare strings and lists by length
        if not isinstance(self.ref, Number):
            self.ref = len(self.ref)


class Shorter(LengthMatcher):
    """Matches values that are shorter (as per ``<`` comparison on ``len``)
    than given value.
    """
    OP = operator.lt

ShorterThan = Shorter


class ShorterOrEqual(LengthMatcher):
    r"""Matches values that are shorter than,
    or equal in ``len``\ gth to (as per ``<=`` operator), given object.
    """
    OP = operator.le

ShorterOrEqualTo = ShorterOrEqual


class Longer(LengthMatcher):
    """Matches values that are longer (as per ``>`` comparison on ``len``)
    than given value.
    """
    OP = operator.gt

LongerThan = Longer


class LongerOrEqual(LengthMatcher):
    r"""Matches values that are longer than,
    or equal in ``len``\ gth to (as per ``>=`` operator), given object.
    """
    OP = operator.ge

LongerOrEqualTo = LongerOrEqual


# Membership tests

class Contains(OperatorMatcher):
    """Matches values that contain (as per the ``in`` operator)
    given reference object.
    """
    OP = operator.contains

    def __repr__(self):
        return "<%r in %s>" % (self.ref, self._get_placeholder_repr())


class In(OperatorMatcher):
    """Matches values that are within the reference object
    (as per the ``in`` operator).
    """
    # There is no ``operator.in_``, so we must define the function ourselves.
    OP = staticmethod(lambda value, ref: value in ref)

    def __repr__(self):
        return "<%s in %r>" % (self._get_placeholder_repr(), self.ref)



"""
Matchers related to functions and other similar callables.
"""
import inspect

from calleee._compat import asyncio
from calleee.base import BaseMatcher


__all__ = [
    'Callable', 'Function', 'GeneratorFunction', 'CoroutineFunction',
]


class FunctionMatcher(BaseMatcher):
    """Matches values of callable types.
    This class shouldn't be used directly.
    """
    def __repr__(self):
        return "<%s>" % (self.__class__.__name__,)


class Callable(FunctionMatcher):
    """Matches any callable object (as per the :func:`callable` function)."""

    def match(self, value):
        return callable(value)


class Function(FunctionMatcher):
    """Matches any Python function."""

    def match(self, value):
        return inspect.isfunction(value)


class GeneratorFunction(FunctionMatcher):
    """Matches a generator function, i.e. one that uses ``yield`` in its body.

    .. note::

        This is distinct from matching a *generator*,
        i.e. an iterable result of calling the generator function,
        or a generator comprehension (``(... for x in ...)``).
        The :class:`~calleee.collections.Generator` matcher
        should be used for those objects instead.
    """
    def match(self, value):
        return inspect.isgeneratorfunction(value)


class CoroutineFunction(FunctionMatcher):
    """Matches a coroutine function.

    A coroutine function is an asynchronous function defined using the
    ``@asyncio.coroutine`` or the ``async def`` syntax.

    These are only available in Python 3.4 and above.
    On previous versions of Python, no object will match this matcher.
    """
    def match(self, value):
        return asyncio and asyncio.iscoroutinefunction(value)



#!/usr/bin/env python
"""
calleee
======

{description}
"""
import ast
from io import open
import os
from setuptools import find_packages, setup
import sys


# Utility functions

def read_tags(filename):
    """Reads values of "magic tags" defined in the given Python file.

    :param filename: Python filename to read the tags from
    :return: Dictionary of tags
    """
    with open(filename) as f:
        ast_tree = ast.parse(f.read(), filename)

    res = {}
    for node in ast.walk(ast_tree):
        if type(node) is not ast.Assign:
            continue

        target = node.targets[0]
        if type(target) is not ast.Name:
            continue
        if not (target.id.startswith('__') and target.id.endswith('__')):
            continue

        name = target.id[2:-2]
        res[name] = ast.literal_eval(node.value)

    return res


def read_requirements(filename='requirements.txt'):
    """Reads the list of requirements from given file.

    :param filename: Filename to read the requirements from.
                     Uses ``'requirements.txt'`` by default.

    :return: Requirements as list of strings
    """
    # allow for some leeway with the argument
    if not filename.startswith('requirements'):
        filename = 'requirements-' + filename
    if not os.path.splitext(filename)[1]:
        filename += '.txt'  # no extension, add default

    def valid_line(line):
        line = line.strip()
        return line and not any(line.startswith(p) for p in ('#', '-'))

    def extract_requirement(line):
        egg_eq = '#egg='
        if egg_eq in line:
            _, requirement = line.split(egg_eq, 1)
            return requirement
        return line

    with open(filename) as f:
        lines = f.readlines()
        return list(map(extract_requirement, filter(valid_line, lines)))


# setup() call

tags = read_tags(os.path.join('calleee', '__init__.py'))

tests_require = read_requirements('test')
if sys.version_info < (2, 7):
    tests_require.extend(read_requirements('test-py26'))
if sys.version_info < (3, 3):
    tests_require.extend(read_requirements('test-py32'))

here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README file
with open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

# TODO: move to setup.cfg
setup(
    name="calleee",
    version=tags['version'],
    description=tags['description'],
    long_description=long_description,
    long_description_content_type="text/x-rst",
    author=tags['author'],
    url="https://github.com/untidy-hair/calleee",
    license=tags['license'],

    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Testing",
    ],

    platforms='any',
    packages=find_packages(exclude=['tests']),

    tests_require=tests_require,
)



"""
Matchers for numbers.
"""
from __future__ import absolute_import

import fractions
import numbers

from calleee._compat import IS_PY3
from calleee.base import BaseMatcher


__all__ = [
    'Number',
    'Complex', 'Real', 'Float', 'Rational', 'Fraction',
    'Integral', 'Integer', 'Int', 'Long',
]


class NumericMatcher(BaseMatcher):
    """Matches some number type.
    This class shouldn't be used directly.
    """
    #: Number class to match.
    #: Must be overridden in subclasses.
    CLASS = None

    def __init__(self):
        assert self.CLASS, "must specify number type to match"

    def match(self, value):
        return isinstance(value, self.CLASS)

    def __repr__(self):
        return "<%s>" % (self.__class__.__name__,)


class Number(NumericMatcher):
    """Matches any number
    (integer, float, complex, custom number types, etc.).
    """
    CLASS = numbers.Number


class Complex(NumericMatcher):
    r"""Matches any complex number.

    This *includes* all real, rational, and integer numbers as well,
    which in Python translates to `float`\ s, fractions, and `int`\ egers.
    """
    CLASS = numbers.Complex


# TODO: consider adding a dedicated matcher for the ``complex`` type;
# right now, though, ``IsA(complex)`` and ``Complex() & ~Real()`` are probably
# acceptable workarounds


class Real(NumericMatcher):
    r"""Matches any real number.

    This includes all rational and integer numbers as well, which in Python
    translates to fractions, and `int`\ egers.
    """
    CLASS = numbers.Real


class Float(NumericMatcher):
    """Matches a floating point number."""

    CLASS = float


class Rational(NumericMatcher):
    r"""Matches a rational number.
    This includes all `int`\ eger numbers as well.
    """
    CLASS = numbers.Rational


class Fraction(NumericMatcher):
    """Matches a fraction object."""

    CLASS = fractions.Fraction


class Integral(NumericMatcher):
    """Matches any integer.
    This ignores the length of integer's internal representation on Python 2.
    """
    CLASS = int if IS_PY3 else (int, long)


class Integer(NumericMatcher):
    """Matches a regular integer.

    On Python 3, there is no distinction between regular and long integer,
    making this matcher and :class:`Long` equivalent.

    On Python 2, this matches the :class:`int` integers exclusively.
    """
    CLASS = int

#: Alias for :class:`Integer`.
Int = Integer


class Long(NumericMatcher):
    """Matches a long integer.

    On Python 3, this is the same as regular integer, making this matcher
    and :class:`Integer` equivalent.

    On Python 2, this matches the :class:`long` integers exclusively.
    """
    CLASS = int if IS_PY3 else long



"""
Type-related matchers.
"""
import inspect

from calleee.base import BaseMatcher


__all__ = [
    'InstanceOf', 'IsA', 'SubclassOf', 'Inherits', 'Type', 'Class',
]


class TypeMatcher(BaseMatcher):
    """Matches an object to a type.
    This class shouldn't be used directly.
    """
    def __init__(self, type_):
        r""":param type\ _: Type to match against"""
        if not isinstance(type_, type):
            raise TypeError("%s requires a type, got %r" % (
                self.__class__.__name__, type_))
        self.type_ = type_

    def __repr__(self):
        return "<%s %r>" % (self.__class__.__name__, self.type_)


# TODO: reverse of this matcher (TypeOf / ClassOf)
class InstanceOf(TypeMatcher):
    """Matches an object that's an instance of given type
    (as per `isinstance`).
    """
    def __init__(self, type_, exact=False):
        r"""
        :param type\ _: Type to match against
        :param exact:

            If True, the match will only succeed if the value type matches
            given ``type_`` exactly.
            Otherwise (the default), a subtype of ``type_`` will also match.
        """
        super(InstanceOf, self).__init__(type_)
        self.exact = exact

    def match(self, value):
        if self.exact:
            return type(value) is self.type_
        else:
            return isinstance(value, self.type_)

IsA = InstanceOf


# TODO: reverse of this matcher (SuperclassOf)
class SubclassOf(TypeMatcher):
    """Matches a class that's a subclass of given type
    (as per `issubclass`).
    """
    def __init__(self, type_, strict=False):
        r"""
        :param type\ _: Type to match against
        :param strict:

            If True, the match if only succeed if the value is a _strict_
            subclass of ``type_`` -- that is, it's not ``type_`` itself.
            Otherwise (the default), any subclass of ``type_`` matches.
        """
        super(SubclassOf, self).__init__(type_)
        self.strict = strict

    def match(self, value):
        if not isinstance(value, type):
            return False
        if not issubclass(value, self.type_):
            return False
        if value is self.type_ and self.strict:
            return False
        return True

Inherits = SubclassOf


class Type(BaseMatcher):
    """Matches any Python type object."""

    def match(self, value):
        return isinstance(value, type)

    def __repr__(self):
        return "<Type>"


class Class(BaseMatcher):
    """Matches a class (but not any other type object)."""

    def match(self, value):
        return inspect.isclass(value)

    def __repr__(self):
        return "<Class>"


# TODO: Module() matcher



"""
Matchers for strings.
"""
import fnmatch
import re

from calleee._compat import IS_PY3, casefold
from calleee.base import BaseMatcher
from calleee.objects import Bytes


__all__ = [
    'Bytes',  # backwards compatibility
    'String', 'Unicode',
    'StartsWith', 'EndsWith',
    'Glob', 'Regex',
]


# String type matchers

class StringTypeMatcher(BaseMatcher):
    """Matches some string type.
    This class shouldn't be used directly.
    """
    #: String class to match.
    #: Must be overridden in subclasses.
    CLASS = None

    # TODO: support of= param, so we can assert what characters
    # the string consists of (e.g. letters, digits as iterables of chars;
    # boolean predicate; or matcher)
    def __init__(self):
        assert self.CLASS, "must specify string type to match"

    def match(self, value):
        return isinstance(value, self.CLASS)

    def __repr__(self):
        return "<%s>" % (self.__class__.__name__,)


class String(StringTypeMatcher):
    """Matches any string.

    | On Python 2, this means either :class:`str` or :class:`unicode` objects.
    | On Python 3, this means :class:`str` objects exclusively.
    """
    CLASS = str if IS_PY3 else basestring


class Unicode(StringTypeMatcher):
    """Matches a Unicode string.

    | On Python 2, this means :class:`unicode` objects exclusively.
    | On Python 3, this means :class:`str` objects exclusively.
    """
    CLASS = str if IS_PY3 else unicode


# Infix matchers

# TODO: generalize for all sequence/collection types

class StartsWith(BaseMatcher):
    """Matches a string starting with given prefix."""

    def __init__(self, prefix):
        self.prefix = prefix

    def match(self, value):
        return value.startswith(self.prefix)

    def __repr__(self):
        return "<StartsWith %r>" % (self.prefix,)


class EndsWith(BaseMatcher):
    """Matches a string ending with given suffix."""

    def __init__(self, suffix):
        self.suffix = suffix

    def match(self, value):
        return value.endswith(self.suffix)

    def __repr__(self):
        return "<EndsWith %r>" % (self.suffix,)


# Pattern matchers

class Glob(BaseMatcher):
    """Matches a string against a Unix shell wildcard pattern.

    See the :mod:`fnmatch` module for more details about those patterns.
    """
    DEFAULT_CASE = 'system'

    #: fnmatch functions that the matchers uses based on case= argument.
    FNMATCH_FUNCTIONS = {
        DEFAULT_CASE: fnmatch.fnmatch,
        True: fnmatch.fnmatchcase,
        False: lambda f, p: fnmatch.fnmatchcase(casefold(f), casefold(p)),
    }

    def __init__(self, pattern, case=None):
        """
        :param pattern: Pattern to match against
        :param case:

            Case sensitivity setting. Possible options:

                * ``'system'`` or ``None``: case sensitvity is system-dependent
                  (this is the default)
                * ``True``: matching is case-sensitive
                * ``False``: matching is case-insensitive
        """
        self.pattern = pattern
        try:
            if case is None:
                case = self.DEFAULT_CASE
            self.fnmatch = self.FNMATCH_FUNCTIONS[case]
        except KeyError:
            raise ValueError("invalid case= argument: %r" % (case,))

    def match(self, value):
        return self.fnmatch(value, self.pattern)

    def __repr__(self):
        return "<Glob %s>" % (self.pattern,)


class Regex(BaseMatcher):
    """Matches a string against a regular expression."""

    REGEX_TYPE = type(re.compile(''))

    def __init__(self, pattern, flags=0):
        """
        :param pattern: Regular expression to match against.
                        It can be given as string,
                        or as a compiled regular expression object
        :param flags: Flags to use with a regular expression passed as string
        """
        if self._is_regex_object(pattern):
            if flags and flags != pattern.flags:
                raise ValueError("conflicting regex flags: %s vs. %s" % (
                    bin(flags), bin(pattern.flags)))
        else:
            pattern = re.compile(pattern, flags)

        self.pattern = pattern

    def _is_regex_object(self, obj):
        return isinstance(obj, self.REGEX_TYPE)

    def match(self, value):
        return self.pattern.match(value)

    def __repr__(self):
        return "<Regex %s>" % (self.pattern.pattern,)


# TODO: matchers for common string formats: Url, Email, IPv4, IPv6



"""
Matchers for various common kinds of objects.
"""
import inspect
import sys

from calleee._compat import asyncio, getargspec
from calleee.base import BaseMatcher


__all__ = ['Bytes', 'Coroutine', 'FileLike']


class ObjectMatcher(BaseMatcher):
    """Base class for object matchers.
    This class shouldn't be used directly.
    """
    def __repr__(self):
        return "<%s>" % (self.__class__.__name__,)


# TODO: Date, DateTime, and Time matchers (with before=/after= params)
# TODO: TimeDelta matcher


class Bytes(ObjectMatcher):
    """Matches a byte array, i.e. the :class:`bytes` type.

    | On Python 2, :class:`bytes` class is identical to :class:`str` class.
    | On Python 3, byte strings are separate class, distinct from :class:`str`.
    """
    def match(self, value):
        return isinstance(value, bytes)


class Coroutine(ObjectMatcher):
    """Matches an asynchronous coroutine.

    A coroutine is a result of an asynchronous function call, where the async
    function has been defined using ``@asyncio.coroutine`` or the ``async def``
    syntax.

    These are only available in Python 3.4 and above.
    On previous versions of Python, no object will match this matcher.
    """
    def match(self, value):
        return asyncio and asyncio.iscoroutine(value)


class FileLike(ObjectMatcher):
    """Matches a file-like object.

    In general, a `file-like object` is an object you can ``read`` data from,
    or ``write`` data to.
    """
    def __init__(self, read=True, write=None):
        """
        :param read:

            Whether only to match objects that do support (``True``)
            or don't support (``False``) reading from them.
            If ``None`` is passed, reading capability is not matched against.

        :param write:

            Whether only to match objects that do support (``True``)
            or don't support (``False``) writing to them.
            If ``None`` is passed, writing capability is not matched against.
        """
        if read is None and write is None:
            raise ValueError("cannot match file-like objects "
                             "that are neither readable nor writable")
        self.read = read if read is None else bool(read)
        self.write = write if write is None else bool(write)

    def match(self, value):
        if self.read is not None:
            if self.read != self._is_readable(value):
                return False
        if self.write is not None:
            if self.write != self._is_writable(value):
                return False
        return True

    def _is_readable(self, obj):
        """Check if the argument is a readable file-like object."""
        try:
            read = getattr(obj, 'read')
        except AttributeError:
            return False
        else:
            return is_method(read, max_arity=1)

    def _is_writable(self, obj):
        """Check if the argument is a writable file-like object."""
        try:
            write = getattr(obj, 'write')
        except AttributeError:
            return False
        else:
            return is_method(write, min_arity=1, max_arity=1)

    def __repr__(self):
        """Return a representation of this matcher."""
        requirements = []
        if self.read is not None:
            requirements.append("read" if self.read else "noread")
        if self.write is not None:
            requirements.append("write" if self.write else "nowrite")
        return "<FileLike %s>" % "(%s)" % ",".join(requirements)


# Utility functions

def is_method(arg, min_arity=None, max_arity=None):
    """Check if argument is a method.

    Optionally, we can also check if minimum or maximum arities
    (number of accepted arguments) match given minimum and/or maximum.
    """
    if not callable(arg):
        return False

    if not any(is_(arg) for is_ in (inspect.ismethod,
                                    inspect.ismethoddescriptor,
                                    inspect.isbuiltin)):
        return False

    try:
        argnames, varargs, kwargs, defaults = getargspec(arg)
    except TypeError:
        # On CPython 2.x, built-in methods of file aren't inspectable,
        # so if it's file.read() or file.write(), we can't tell it for sure.
        # Given how this check is being used, assuming the best is probably
        # all we can do here.
        return True
    else:
        if argnames and argnames[0] == 'self':
            argnames = argnames[1:]

    if min_arity is not None:
        actual_min_arity = len(argnames) - len(defaults or ())
        assert actual_min_arity >= 0, (
            "Minimum arity of %r found to be negative (got %s)!" % (
                arg, actual_min_arity))
        if int(min_arity) != actual_min_arity:
            return False

    if max_arity is not None:
        actual_max_arity = sys.maxsize if varargs or kwargs else len(argnames)
        if int(max_arity) != actual_max_arity:
            return False

    return True
